﻿using System;
using System.Globalization;
using System.IO;
using System.Web;
using System.Linq;
using System.Collections.Generic;
using Dynamicweb.Rendering;
using Dynamicweb.Ecommerce.LiveIntegration.Configuration;
using Dynamicweb.Core.Helpers;
using Dynamicweb.Mailing;

namespace Dynamicweb.Ecommerce.LiveIntegration
{    
    public enum ErrorLevel 
    { 
        //response and request content
        DebugInfo, 
        //web service connection errors
        ConnectionError,
        //response is empty or response format is broken
        ResponseError,
        //general error
        Error,
        //email send is used for getting errors appeared after last email sending
        EmailSend
    };

    public enum NotificationFrequency
    {
        Never = 0,
        Minute = 1,
        FiveMinutes = 5,
        TenMinutes = 10,
        FifteenMinutes = 15,
        ThirtyMinutes = 30,
        OneHour = 60,
        TwoHours = 120,
        ThreeHours = 180,
        SixHours = 360,
        TwelveHours = 720,
        OneDay = 1440
    }

    public class Logger
	{
        private static readonly Logger instance = new Logger();

		private static readonly object syncLock = new object();
		private static readonly CultureInfo _ci = new CultureInfo("en-US");
        private readonly string _logFile;        

		private Logger()
		{
            if (HttpContext.Current != null)
            {
                _logFile = HttpContext.Current.Server.MapPath("/Files/System/Log/LiveIntegration/" + Configuration.Constants.AddInName + ".log");
            }
		}

        public static Logger Instance
        {
            get
            {
                return instance;
            }
        }

        private bool LogConnectionErrors
        {
            get
            { 
                return Convert.ToBoolean(Helpers.GetConfigValue(Configuration.Constants.ErrorHandlingConfiguration.Path, Configuration.Constants.ErrorHandlingConfiguration.LogConnectionErrors));
            }
        }

        private bool LogResponseErrors
        {
            get
            {
                return Convert.ToBoolean(Helpers.GetConfigValue(Configuration.Constants.ErrorHandlingConfiguration.Path, Configuration.Constants.ErrorHandlingConfiguration.LogResponseErrors));
            }
        }

        private bool LogGeneralErrors
        {
            get
            {
                return Convert.ToBoolean(Helpers.GetConfigValue(Configuration.Constants.ErrorHandlingConfiguration.Path, Configuration.Constants.ErrorHandlingConfiguration.LogGeneralErrors));
            }
        }

        private bool LogDebugInfo
        {
            get
            {
                return Convert.ToBoolean(Helpers.GetConfigValue(Configuration.Constants.ErrorHandlingConfiguration.Path, Configuration.Constants.ErrorHandlingConfiguration.LogDebugInfo));
            }
        }

        public void Log(ErrorLevel errorLevel, string logLine)
        {
            if(!string.IsNullOrEmpty(_logFile))
            {
                bool isAllowedAddToLog = IsAllowedAddToLog(errorLevel);
                lock(syncLock)
                {
                    if(isAllowedAddToLog)
                    {
                        TruncateLogFile();

                        logLine = DateTime.Now.ToString(_ci) + ": " + errorLevel.ToString() + ": " + logLine + System.Environment.NewLine;
                        TextFileHelper.WriteTextFile(logLine, _logFile, true, System.Text.Encoding.UTF8);
                    }
                }
                if(errorLevel != ErrorLevel.DebugInfo && errorLevel != ErrorLevel.EmailSend)
                {
                    SendLog(logLine, isAllowedAddToLog);
                }
            }
        }

        private void TruncateLogFile()
        {
            FileInfo fi = new FileInfo(_logFile);

            try
            {
                int maxSize;

                if(!int.TryParse(Helpers.GetConfigValue(Configuration.Constants.ErrorHandlingConfiguration.Path, Configuration.Constants.ErrorHandlingConfiguration.LogMaxSize), out maxSize))
                    maxSize = 10;

                if(maxSize > 100)
                    maxSize = 100;

                if(fi.Exists && fi.Length > maxSize * 1024 * 1024 * 1.1) // +10%
                {
                    string folder = Path.GetDirectoryName(_logFile);
                    string newFile = Path.Combine(folder, Path.GetRandomFileName());

                    using(FileStream ws = new FileStream(newFile, FileMode.CreateNew))
                    {
                        using(FileStream rs = new FileStream(_logFile, FileMode.Open))
                        {
                            rs.Seek((long)(fi.Length - maxSize * 1024 * 1024 * 0.5), SeekOrigin.Begin); // -50%
                            int currChar;

                            while((currChar = rs.ReadByte()) != -1)
                            {
                                if(currChar != '\r')
                                    continue;

                                currChar = rs.ReadByte();

                                if(currChar == -1 || currChar == '\n')
                                    break;
                            }

                            rs.CopyTo(ws, 1048576);
                        }

                        ws.Flush(true);
                    }

                    File.Delete(_logFile);
                    File.Move(newFile, _logFile);
                }
            }
            catch(Exception ex)
            {
                string logLine = DateTime.Now.ToString(_ci) + ": " + ErrorLevel.Error + ": Cannot truncate log file." + System.Environment.NewLine;
                TextFileHelper.WriteTextFile(logLine, _logFile, true, System.Text.Encoding.UTF8);
            }
        }

        private void SendLog(string lastError, bool isLastErrorInLog)
        {
            string frequencySettings = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationSendingFrequency);            
            if (!string.IsNullOrEmpty(frequencySettings))
            {
                NotificationFrequency frequency = Helpers.GetEnumValueFromString<NotificationFrequency>(frequencySettings, NotificationFrequency.Never);
                if (frequency != NotificationFrequency.Never)
                {                    
                    //get last time when the email was sent
                    string strLastTimeSend = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationLastSendTime);
                    bool emailSend = false;
                    if (string.IsNullOrEmpty(strLastTimeSend))
                    {
                        emailSend = SendMail(lastError);
                    }
                    else
                    {
                        //send email if the frequency interval already passed
                        DateTime lastTimeSend;                                                 
                        if (DateTime.TryParse(strLastTimeSend, _ci, DateTimeStyles.None, out lastTimeSend) &&
                            (TimeSpan)DateTime.Now.Subtract(lastTimeSend) >= TimeSpan.FromMinutes((double)frequency))
                        {
                            if (!isLastErrorInLog)
                            {
                                emailSend = SendMail(GetLastLogData() + lastError);
                            }
                            else
                            {
                                emailSend = SendMail(GetLastLogData());
                            }                            
                        }
                    }
                    if (emailSend)
                    {
                        Log(ErrorLevel.EmailSend, "Send e-mail with errors");//used for getting the last errors appeared for the future email
                        Dictionary<string, string> values = ConfigurationHandler.GetConfig(Configuration.Constants.NotificationSettings.Path);
                        if (values.ContainsKey(Configuration.Constants.NotificationSettings.NotificationLastSendTime))
                        {
                            values[Configuration.Constants.NotificationSettings.NotificationLastSendTime] = DateTime.Now.ToString(_ci);
                        }
                        else
                        {
                            values.Add(Configuration.Constants.NotificationSettings.NotificationLastSendTime, DateTime.Now.ToString(_ci));
                        }
                        Configuration.ConfigurationHandler.SetConfig(Configuration.Constants.NotificationSettings.Path, values);
                    }
                }                
            }
        }

        /// <summary>
        /// Gets appeared errors to send from the last email send
        /// </summary>
        /// <returns></returns>
        public string GetLastLogData()
        {
            string result = "";

            lock (syncLock)
            {                
                foreach (var line in File.ReadLines(_logFile, System.Text.Encoding.UTF8).Reverse())
                {
                    if (line.Contains(ErrorLevel.DebugInfo.ToString()))
                    {
                        continue;
                    }
                    else if (!line.Contains(ErrorLevel.EmailSend.ToString()))
                    {
                        result += line + "<br>";
                    }
                    else
                    {
                        break;
                    }
                }                
            }
            return result;
        }

        /// <summary>
        /// Send mail according to configuration 
        /// </summary>
        /// <param name="message">error/success message</param>
        public static bool SendMail(string message)
        {
            bool result = false;
            string notificationTemplate = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationTemplate);
            string notificationEmail = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationEmail);

            if (!string.IsNullOrEmpty(message) && !string.IsNullOrEmpty(notificationTemplate) && StringHelper.IsValidEmailAddress(notificationEmail))
            {
                Template templateInstance = new Template(notificationTemplate);
                if (templateInstance != null)
                {
                    templateInstance.SetTag("Ecom:LiveIntegration.AddInName", Configuration.Constants.AddInName);
                    templateInstance.SetTag("Ecom:LiveIntegration.ErrorMessage", message);

                    string notificationEmailSubject = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationEmailSubject);
                    string notificationEmailSenderEmail = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationEmailSenderEmail);
                    string notificationEmailSenderName = Helpers.GetConfigValue(Configuration.Constants.NotificationSettings.Path, Configuration.Constants.NotificationSettings.NotificationEmailSenderName);

                    var mail = new System.Net.Mail.MailMessage
                    {
                        IsBodyHtml = true,
                        Subject = notificationEmailSubject,
                        SubjectEncoding = System.Text.Encoding.UTF8,
                        From = new System.Net.Mail.MailAddress(notificationEmailSenderEmail, notificationEmailSenderName, System.Text.Encoding.UTF8)
                    };
                    //Set parameters for MailMessage
                    mail.To.Add(notificationEmail);
                    mail.BodyEncoding = System.Text.Encoding.UTF8;
                    //Render Template and set as Body
                    mail.Body = templateInstance.Output();
                    //Send mail
                    EmailHandler.Send(mail);
                    result = true;
                }
            }
            return result;
        }

        private bool IsAllowedAddToLog(ErrorLevel level)
        {
            bool result = false;
            switch (level)
            {
                case ErrorLevel.ConnectionError:
                    result = LogConnectionErrors;
                    break;
                case ErrorLevel.ResponseError:
                    result = LogResponseErrors;
                    break;
                case ErrorLevel.Error:
                    result = LogGeneralErrors;
                    break;
                case ErrorLevel.DebugInfo:
                    result = LogDebugInfo;
                    break;
                case ErrorLevel.EmailSend:
                    result = true;
                    break;
            }            
            return result;
        }
    }
}
